@@ -137,7 +137,9 @@ impl<B: Bound> Constraint<B> {
 }
 
 /// Represents a zone with a specific *[bound type][Bound]*.
-pub trait Zone<B: Bound> {
+pub trait Zone {
+    type Bound: Bound;
+
     /// Constructs a new zone without any constraints besides clocks being positive.
     fn new_unconstrained(num_variables: usize) -> Self;
     /// Constructs a new zone where all clocks are set to zero.
@@ -146,7 +148,7 @@ pub trait Zone<B: Bound> {
     /// Constructs a new zone from the given iterable of [constraints][Constraint].
     fn with_constraints<U>(num_variables: usize, constraints: U) -> Self
     where
-        U: IntoIterator<Item = Constraint<B>>;
+        U: IntoIterator<Item = Constraint<Self::Bound>>;
 
     /// Returns the number of clock variables of this zone.
     fn num_variables(&self) -> usize;
@@ -158,17 +160,17 @@ pub trait Zone<B: Bound> {
     fn num_clocks(&self) -> usize;
 
     /// Retrieves the difference bound for `left - right`.
-    fn get_bound(&self, left: impl AnyClock, right: impl AnyClock) -> &B;
+    fn get_bound(&self, left: impl AnyClock, right: impl AnyClock) -> &Self::Bound;
 
     /// Checks whether the zone is empty.
     fn is_empty(&self) -> bool;
 
     /// Adds a [constraint][Constraint] to the zone.
-    fn add_constraint(&mut self, constraint: Constraint<B>);
+    fn add_constraint(&mut self, constraint: Constraint<Self::Bound>);
     /// Adds all [constraints][Constraint] from the given iterable to the zone.
     fn add_constraints<U>(&mut self, constraints: U)
     where
-        U: IntoIterator<Item = Constraint<B>>;
+        U: IntoIterator<Item = Constraint<Self::Bound>>;
 
     /// Intersects two zones.
     fn intersect(&mut self, other: &Self);
@@ -183,18 +185,18 @@ pub trait Zone<B: Bound> {
     fn past(&mut self);
 
     /// Resets the given clock variable to the specified constant.
-    fn reset(&mut self, clock: Variable, value: B::Constant);
+    fn reset(&mut self, clock: Variable, value: <Self::Bound as Bound>::Constant);
 
     /// Checks whether the value of the given clock is unbounded.
     fn is_unbounded(&self, clock: impl AnyClock) -> bool;
 
     /// Returns the upper bound for the value of the given clock.
-    fn get_upper_bound(&self, clock: impl AnyClock) -> Option<B::Constant>;
+    fn get_upper_bound(&self, clock: impl AnyClock) -> Option<<Self::Bound as Bound>::Constant>;
     /// Returns the lower bound for the value of the given clock.
-    fn get_lower_bound(&self, clock: impl AnyClock) -> Option<B::Constant>;
+    fn get_lower_bound(&self, clock: impl AnyClock) -> Option<<Self::Bound as Bound>::Constant>;
 
     /// Checks whether the given constraint is satisfied by the zone.
-    fn is_satisfied(&self, constraint: Constraint<B>) -> bool;
+    fn is_satisfied(&self, constraint: Constraint<Self::Bound>) -> bool;
 
     /// Checks whether the zone includes the other zone.
     fn includes(&self, other: &Self) -> bool;
@@ -207,13 +209,13 @@ pub trait Zone<B: Bound> {
 
 /// Returns an iterator over the clocks of a zone.
 #[inline(always)]
-pub fn clocks<Z: Zone<B>, B: Bound>(zone: &Z) -> impl Iterator<Item = Clock> {
+pub fn clocks<Z: Zone>(zone: &Z) -> impl Iterator<Item = Clock> {
     (0..zone.num_clocks()).map(Clock)
 }
 
 /// Returns an iterator over the variables of a zone.
 #[inline(always)]
-pub fn variables<Z: Zone<B>, B: Bound>(zone: &Z) -> impl Iterator<Item = Variable> {
+pub fn variables<Z: Zone>(zone: &Z) -> impl Iterator<Item = Variable> {
     (1..zone.num_clocks()).map(Variable)
 }
 
@@ -234,15 +236,11 @@ impl<B: Bound, L: Layout<B>> Dbm<B, L> {
     fn new(num_variables: usize, default: B) -> Self {
         let dimension = num_variables + 1;
         let mut layout = L::new(num_variables, default);
-        unsafe {
-            // This is safe because we know that the clock indices are
-            // within the dimensions of the DBM.
-            layout.set_unchecked(Clock::ZERO, Clock::ZERO, B::le_zero());
-            for index in 1..dimension {
-                let clock = Clock::from_index(index);
-                layout.set_unchecked(Clock::ZERO, clock, B::le_zero());
-                layout.set_unchecked(clock, clock, B::le_zero());
-            }
+        layout.set(Clock::ZERO, Clock::ZERO, B::le_zero());
+        for index in 1..dimension {
+            let clock = Clock::from_index(index);
+            layout.set(Clock::ZERO, clock, B::le_zero());
+            layout.set(clock, clock, B::le_zero());
         }
         Dbm {
             dimension,
@@ -251,20 +249,9 @@ impl<B: Bound, L: Layout<B>> Dbm<B, L> {
         }
     }
 
-    #[inline(always)]
-    fn check_clock(&self, clock: impl AnyClock) {
-        assert!(clock.into_index() < self.dimension, "invalid clock index");
-    }
-
     /// Canonicalize the DBM given that a particular clock has been touched.
-    ///
-    /// This method is unsafe because it does not check wether the index of the
-    /// touched clock is within the dimensions of the DBM.
-    ///
-    /// Providing a clock which is not within the dimensions
-    /// of the DBM is undefined behavior.
     #[inline(always)]
-    unsafe fn canonicalize_touched(&mut self, touched: Clock) {
+    fn canonicalize_touched(&mut self, touched: Clock) {
         for index in 0..self.dimension {
             let left = Clock::from_index(index);
             for index in 0..self.dimension {
@@ -272,10 +259,10 @@ impl<B: Bound, L: Layout<B>> Dbm<B, L> {
                 let bound = self
                     .layout
                     .get(left, touched)
-                    .add(self.layout.get_unchecked(touched, right))
+                    .add(self.layout.get(touched, right))
                     .expect("overflow while adding bounds");
                 if bound.is_tighter_than(self.layout.get(left, right)) {
-                    self.layout.set_unchecked(left, right, bound);
+                    self.layout.set(left, right, bound);
                 }
             }
         }
@@ -284,14 +271,14 @@ impl<B: Bound, L: Layout<B>> Dbm<B, L> {
     fn canonicalize(&mut self) {
         for index in 0..self.dimension {
             let touched = Clock::from_index(index);
-            // This is safe because the touched clock is within
-            // the dimensions of the DBM.
-            unsafe { self.canonicalize_touched(touched) };
+            self.canonicalize_touched(touched);
         }
     }
 }
 
-impl<B: Bound, L: Layout<B>> Zone<B> for Dbm<B, L> {
+impl<B: Bound, L: Layout<B>> Zone for Dbm<B, L> {
+    type Bound = B;
+
     fn new_unconstrained(num_variables: usize) -> Self {
         Dbm::new(num_variables, B::unbounded())
     }
@@ -329,23 +316,18 @@ impl<B: Bound, L: Layout<B>> Zone<B> for Dbm<B, L> {
 
     #[inline(always)]
     fn is_empty(&self) -> bool {
-        // This is safe because the zero clock is always present.
-        unsafe { self.layout.get_unchecked(Clock::ZERO, Clock::ZERO) }
+        self.layout
+            .get(Clock::ZERO, Clock::ZERO)
             .is_tighter_than(&B::le_zero())
     }
 
     fn add_constraint(&mut self, constraint: Constraint<B>) {
-        self.check_clock(constraint.left);
-        self.check_clock(constraint.right);
         let bound = self.layout.get(constraint.left, constraint.right);
         if constraint.bound.is_tighter_than(&bound) {
-            unsafe {
-                // This is safe because we just checked the clocks.
-                self.layout
-                    .set_unchecked(constraint.left, constraint.right, constraint.bound);
-                self.canonicalize_touched(constraint.left);
-                self.canonicalize_touched(constraint.right);
-            }
+            self.layout
+                .set(constraint.left, constraint.right, constraint.bound);
+            self.canonicalize_touched(constraint.left);
+            self.canonicalize_touched(constraint.right);
         }
     }
 
@@ -367,20 +349,13 @@ impl<B: Bound, L: Layout<B>> Zone<B> for Dbm<B, L> {
             let left = Clock::from_index(index);
             for index in 0..self.dimension {
                 let right = Clock::from_index(index);
-                unsafe {
-                    // This is safe because we know that the clock indices are
-                    // within the dimensions of the DBMs.
-                    if other
-                        .layout
-                        .get_unchecked(left, right)
-                        .is_tighter_than(&self.layout.get(left, right))
-                    {
-                        self.layout.set_unchecked(
-                            left,
-                            right,
-                            other.layout.get(left, right).clone(),
-                        );
-                    }
+                if other
+                    .layout
+                    .get(left, right)
+                    .is_tighter_than(&self.layout.get(left, right))
+                {
+                    self.layout
+                        .set(left, right, other.layout.get(left, right).clone());
                 }
             }
         }
@@ -390,9 +365,7 @@ impl<B: Bound, L: Layout<B>> Zone<B> for Dbm<B, L> {
     fn future(&mut self) {
         for index in 1..self.dimension {
             let left = Clock::from_index(index);
-            // This is safe because we know that the clock index is
-            // within the dimensions of the DBMs.
-            unsafe { self.layout.set_unchecked(left, Clock::ZERO, B::unbounded()) };
+            self.layout.set(left, Clock::ZERO, B::unbounded());
         }
     }
     fn past(&mut self) {
@@ -416,29 +389,24 @@ impl<B: Bound, L: Layout<B>> Zone<B> for Dbm<B, L> {
     fn reset(&mut self, clock: Variable, value: B::Constant) {
         let le_pos_value = B::new_le(value.clone());
         let le_neg_value = B::new_le(value.checked_neg().unwrap());
-        self.check_clock(clock);
         for index in 0..self.dimension {
             let other = Clock::from_index(index);
-            // This is safe because the index of `other` is guaranteed to be
-            // within the dimensions of the DBM and we checked `clock`.
-            unsafe {
-                self.layout.set_unchecked(
-                    clock,
-                    other,
-                    self.layout
-                        .get_unchecked(Clock::ZERO, other)
-                        .add(&le_pos_value)
-                        .unwrap(),
-                );
-                self.layout.set(
-                    other,
-                    clock,
-                    self.layout
-                        .get_unchecked(other, Clock::ZERO)
-                        .add(&le_neg_value)
-                        .unwrap(),
-                );
-            }
+            self.layout.set(
+                clock,
+                other,
+                self.layout
+                    .get(Clock::ZERO, other)
+                    .add(&le_pos_value)
+                    .unwrap(),
+            );
+            self.layout.set(
+                other,
+                clock,
+                self.layout
+                    .get(other, Clock::ZERO)
+                    .add(&le_neg_value)
+                    .unwrap(),
+            );
         }
     }
 
@@ -471,16 +439,12 @@ impl<B: Bound, L: Layout<B>> Zone<B> for Dbm<B, L> {
             let left = Clock::from_index(index);
             for index in 0..self.dimension {
                 let right = Clock::from_index(index);
-                unsafe {
-                    // This is safe because we know that the indices are within
-                    // the dimensions of the DBM.
-                    if self
-                        .layout
-                        .get_unchecked(left, right)
-                        .is_tighter_than(other.layout.get(left, right))
-                    {
-                        return false;
-                    }
+                if self
+                    .layout
+                    .get(left, right)
+                    .is_tighter_than(other.layout.get(left, right))
+                {
+                    return false;
                 }
             }
         }
@@ -493,12 +457,8 @@ impl<B: Bound, L: Layout<B>> Zone<B> for Dbm<B, L> {
             let left = Clock::from_index(index);
             for index in 0..min(self.dimension, other.dimension) {
                 let right = Clock::from_index(index);
-                // This is safe because we know that the clock indices
-                // are within the dimensions of both DBMs.
-                unsafe {
-                    let bound = self.layout.get_unchecked(left, right);
-                    other.layout.set_unchecked(left, right, bound.clone());
-                }
+                let bound = self.layout.get(left, right);
+                other.layout.set(left, right, bound.clone());
             }
         }
         other.canonicalize();
